from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_ttk as ttk

import re

"""
Defines helper classes to build forms
"""


class FormField(object):
    """
    Base class for form fields objects. Most methods should be overridden.
    """
    def __init__(self, text='', tags=None):
        """
        Constructor
        """
        self.text = text
        self.tags = tags or []

    def reset(self):
        '''
        Resets internal formField state
        '''
        raise NotImplementedError

    def from_string(self, s):
        """
        Sets field value from argument string
        """
        raise NotImplementedError

    def to_string(self):
        """
        Translates field value to string
        """
        raise NotImplementedError

    def get_widget(self, master):
        """
        Creates a new "entry" widget
        """
        raise NotImplementedError

    def get_widgets(self, master):
        """
        Returns a pair consisting of a label and an "entry" widget
        """
        return tk.Label(master, text=self.text), self.get_widget(master)

    def disables(self):
        """
        Returns a list of tags that are "disabled" by the current FormField
        """
        return []

    def enables(self):
        """
        Returns a list of tags that are "enabled" by the current FormField
        """
        return []


class TextField(FormField):
    """
    A FormField holding a text (a tk.Entry)
    """
    def __init__(self, text, default_value='', validate_command=None,
                 password_mode=False, tags=None):
        super(TextField, self).__init__(text, tags=tags)

        self.var = tk.StringVar()
        self.default_value = default_value
        self.reset()
        self.validate_command = validate_command
        self.password_mode = password_mode

    def reset(self):
        self.var.set(self.default_value)

    def from_string(self, s):
        self.var.set(s)

    def to_string(self):
        return self.var.get()

    def get_widget(self, master):
        show = ''
        if self.password_mode:
            show = '*'

        entry = tk.Entry(master, textvariable=self.var, show=show)

        if self.validate_command is not None:
            vcmd = (master.register(self.validate_command), '%P')
            entry.config(validate='key', validatecommand=vcmd)

        return entry


class HostnameField(TextField):
    """
    A TextField accepting host names only
    """
    __hostname_re = re.compile('^([_\-\d\w]+(\.)?)*$')

    def __init__(self, text, default_value='', tags=None):
        super(HostnameField, self).__init__(text, default_value, self.validate,
                                            tags=tags)

    def validate(self, v):
        ok = self.__hostname_re.match(v) is not None

        return ok


class PasswordField(TextField):
    """
    A TextField for a password. The tk.Entry displays stars instead of the
    current text value
    """
    def __init__(self, text, encode=lambda s: s, decode=lambda s: s,
                 return_cb=None, tags=None):
        super(PasswordField, self).__init__(text, password_mode=True,
                                            tags=tags)

        self.encode = encode
        self.decode = decode
        self.return_cb = return_cb

    def from_string(self, s):
        self.var.set(self.decode(s))

    def to_string(self):
        return self.encode(self.var.get())

    def get_widget(self, master):
        entry = super(PasswordField, self).get_widget(master)

        if self.return_cb is not None:
            entry.bind('<Return>', lambda e: self.return_cb())

        return entry


class IntegerField(TextField):
    """
    A TextField accepting only integer litterals
    """
    def validate(self, v):
        try:
            iv = int(v)
            return True
        except:
            pass

        return False

    def __init__(self, text, default_value=0, tags=None):
        super(IntegerField, self).__init__(text, default_value, self.validate,
                                           tags=tags)


class BooleanField(FormField):
    """
    A FormField holding a bool, manifested by a tk.Checkbutton
    """
    def __init__(self, text, default_value=False, disables_tags=None,
                 enables_tags=None, tags=None):
        super(BooleanField, self).__init__(text, tags=tags)

        self.text = text
        self.var = tk.BooleanVar()
        self.default_value = default_value
        self.reset()
        self.disables_tags = disables_tags or []
        self.enables_tags = enables_tags or []

    def reset(self):
        self.var.set(self.default_value)

    def from_string(self, s):
        self.var.set(s)

    def to_string(self):
        return str(self.var.get())

    def get_widget(self, master):
        def changed(*args):
            master.disenables_changed(self)

        return tk.Checkbutton(master, text=self.text, variable=self.var,
                              onvalue=True, offvalue=False,
                              command=changed)

    def get_widgets(self, master):
        return (self.get_widget(master), )

    def disables(self):
        return self.disables_tags

    def enables(self):
        return self.enables_tags

    def disables_state(self):
        if self.var.get():
            return tk.DISABLED

        return tk.NORMAL

    def enables_state(self):
        if self.var.get():
            return tk.NORMAL

        return tk.DISABLED


class FieldContainer(tk.Frame, object):
    """
    A tk.Frame sub-class specialized to propagate state changes to its children
    """
    def config(self, **options):
        frame_options = options.copy()

        if 'state' in frame_options:
            v = frame_options['state']
            del frame_options['state']

            for slave in self.slaves():
                slave.config(state=v)

        tk.Frame.config(self, **frame_options)


class RadioChoiceField(FormField):
    """
    A FormField allowing to choose between several fixed values, manifested by
    tk.Radiobuttons
    """
    def __init__(self, text, values, default_value=None, vertical=True,
                 tags=None):
        super(RadioChoiceField, self).__init__(text, tags=tags)

        self.values = values
        self.var = tk.StringVar()
        if default_value is None:
            default_value = values[0]

        self.default_value = default_value
        self.reset()

        self.side = tk.TOP
        if not vertical:
            self.side = tk.LEFT

    def reset(self):
        self.var.set(self.default_value)


    def from_string(self, s):
        self.var.set(s)

    def to_string(self):
        return self.var.get()

    def get_widget(self, master):
        frame = FieldContainer(master)

        for v in self.values:
            rb = tk.Radiobutton(frame, text=v, value=v, variable=self.var)
            rb.pack(side=self.side)

        return frame


class ComboboxChoiceField(FormField):
    """
    A FormField allowing to choose between several fixed values, manifested by
    a ttk.Combobox
    """
    def __init__(self, text, values, default_value=None, tags=None):
        super(ComboboxChoiceField, self).__init__(text, tags=tags)

        self.values = values
        self.var = tk.StringVar()
        if default_value is None:
            default_value = values[0]

        self.default_value = default_value
        self.reset()

    def reset(self):
        self.var.set(self.default_value)

    def from_string(self, s):
        self.var.set(s)

    def to_string(self):
        return self.var.get()

    def get_widget(self, master):
        return ttk.Combobox(master, values=self.values, textvariable=self.var)


class FormFrame(tk.Frame, object):
    """
    A tk.Frame fr holding a form (i.e. a list of FormField generated widget
    pairs)
    """
    class TagTargets(list):
        def __init__(self):
            self.state = tk.DISABLED

    def __init__(self, master):
        tk.Frame.__init__(self, master)

        self.disablers = {}
        self.enablers = {}
        self.tag_dict = {}

    def grid_fields(self, fieldlist, focus_first=True):
        disablers = {}
        enablers = {}
        tag_dict = {}
        i = 0
        for field in fieldlist:
            tags = field.tags
            disables = field.disables()
            if disables:
                disablers[field] = disables
            enables = field.enables()
            if enables:
                enablers[field] = enables

            widgets = field.get_widgets(self)
            j = 0
            for w in widgets:
                for t in tags:
                    v = tag_dict.get(t, self.TagTargets())
                    v.append(w)
                    tag_dict[t] = v

                w.grid(row=i, column=j)
                j += 1
            if i == 0 and focus_first:
                widgets[-1].focus_set()
            i += 1

        self.disablers.update(disablers)
        self.enablers.update(enablers)
        self.tag_dict.update(tag_dict)

        for disabler in disablers:
            self.set_disables(disabler)

        for enabler in enablers:
            self.set_enables(enabler)

    def set_disables(self, field):
        tags = self.disablers[field]

        state = field.disables_state()

        for tag in tags:
            tag_targets = self.tag_dict[tag]
            tag_targets.state = state
            for w in tag_targets:
                state = tk.NORMAL
                for tt in self.tag_dict.values():
                    if tt.state == tk.DISABLED and w in tt:
                        state = tk.DISABLED
                    w.config(state=state)

    def set_enables(self, field):
        tags = self.enablers[field]

        state = field.enables_state()

        for tag in tags:
            tag_targets = self.tag_dict[tag]
            tag_targets.state = state
            for w in tag_targets:
                state = tk.NORMAL
                for tt in self.tag_dict.values():
                    if tt.state == tk.DISABLED and w in tt:
                        state = tk.DISABLED
                    w.config(state=state)

    def disenables_changed(self, field):
        if field in self.enablers:
            self.set_enables(field)
        if field in self.disablers:
            self.set_disables(field)

if __name__ == '__main__':
    master = tk.Tk()

    hf = HostnameField('host:')

    bf = BooleanField('toto', disables_tags=['toto'])

    pf = PasswordField('pwd:', tags=['toto'])

    if_ = IntegerField('integer:', 12, tags=['toto'])

    rbc = RadioChoiceField('radiochoice:', ['one', 'two', 'three'],
                           vertical=False, tags=['toto'])

    cbcf = ComboboxChoiceField('comboboxchoice:', ['yi', 'er', 'san'],
                               tags=['toto'])

    ff = FormFrame(master)
    ff.grid_fields([hf, bf, pf, if_, rbc, cbcf])
    ff.pack()
    master.mainloop()



import os
import shutil
from datetime import datetime

from six import print_


class Catalog(object):
    """
    Base class for catalog objects. All methods have to be overridden by
    sub-classes.
    """
    def lstat(self, path):
        """
        Returns a dictionary of informations about specified path.
        """
        raise NotImplementedError

    def listdir(self, path):
        """
        Returns directory contents (only filenames, not absolute paths).
        """
        raise NotImplementedError

    def isdir(self, path):
        """
        Returns wether the specified path is a directory.
        """
        raise NotImplementedError

    def join(self, *args):
        """
        Performs the equivalent of os.path.join() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def splitname(self, path):
        """
        Performs the equivalent of os.path.split() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def dirname(self, path):
        """
        Performs the equivalent of os.path.dirname() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def basename(self, path):
        """
        Performs the equivalent of os.path.basename() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def download_files(self, pathlist, destdir):
        """
        Downloads files from the catalog pathlist to local destdir.
        """
        raise NotImplementedError

    def download_directories(self, pathlist, destdir):
        """
        Downloads directories contents to local destdir.
        """
        raise NotImplementedError

    def upload_files(self, files, path):
        """
        Uploads local files to catalog destination path (a directory).
        """
        raise NotImplementedError

    def upload_directories(self, dirs, path):
        """
        Uploads local directories content to catalog destination path (a
        directory).
        """
        raise NotImplementedError

    def delete_files(self, files):
        """
        Deletes catalog files.
        """
        raise NotImplementedError

    def delete_directories(self, directories):
        """
        Recusively deletes catalog directories.
        """
        raise NotImplementedError

    def mkdir(self, path):
        """
        Creates a new catalog directory.
        """
        raise NotImplementedError

    def directory_properties(self, path):
        """
        Returns a dictionary of properties for a directory path
        """
        raise NotImplementedError

    def file_properties(self, path):
        """
        Returns a dictionary of properties for a file path
        """
        raise NotImplementedError

    @classmethod
    def config_fields(cls):
        """
        Returns a dictionary of form.FormField objects for catalog object
        configuration.
        """
        raise NotImplementedError

class OSCatalog(Catalog):
    """
    Presents contents from local filesystem. Useful for debugging.
    """
    def lstat(self, path):
        stats = os.lstat(path)

        nreplicas = '1'
        if self.isdir(path):
            nreplicas = ''
        ret = {
            'user': stats.st_uid,
            'size': stats.st_size,
            'mtime': datetime.fromtimestamp(stats.st_mtime),
            'nreplicas': nreplicas,
        }

        return ret

    def listdir(self, path):
        return os.listdir(path)

    def isdir(self, path):
        return os.path.isdir(path)

    def join(self, *args):
        return os.path.join(*args)

    def splitname(self, path):
        return os.path.split(path)

    def dirname(self, path):
        return os.path.dirname(path)

    def basename(self, path):
        return os.path.basename(path)

    def download_files(self, pathlist, destdir):
        number = len(pathlist)
        i = 0
        for path in pathlist:
            shutil.copy2(path, destdir)
            i += 1
            yield i, number

    def download_directories(self, pathlist, destdir):
        number = len(pathlist)
        i = 0
        for path in pathlist:
            # shutil.copytree needs a fresh new destination directory
            ddir = os.path.join(destdir, os.path.basename(path))
            if os.path.exists(ddir):
                shutil.rmtree(ddir)
            shutil.copytree(path, ddir)
            i += 1
            yield i, number

    def upload_files(self, files, path):
        number = len(files)
        i = 0
        for f in files:
            shutil.copy2(f, path)
            i += 1
            yield i, number

    def upload_directories(self, dirs, path):
        number = len(dirs)
        i = 0
        for d in dirs:
            # shutil.copytree needs a fresh new destination directory
            ddir = os.path.join(path, os.path.basename(d))
            if os.path.exists(ddir):
                shutil.rmtree(ddir)

            shutil.copytree(d, ddir)
            i += 1
            yield i, number

    def delete_files(self, files):
        number = len(files)
        i = 0
        for f in files:
            os.unlink(f)
            i += 1
            yield i, number


    def delete_directories(self, directories):
        number = len(directories)
        i = 0
        for d in directories:
            shutil.rmtree(d)
            i += 1
            yield i, number

    def mkdir(self, path):
        os.mkdir(path)

    def directory_properties(self, path):
        return {}

    def file_properties(self, path):
        return {}

    @classmethod
    def config_fields(cls):
        return {}





from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_tksimpledialog as tksimpledialog
from six.moves import tkinter_ttk as ttk

import collections
import copy

class ColumnDef(object):
    def __init__(self, name, text, anchor='w', form_field=None):
        self.name = name
        self.text = text
        self.anchor = anchor
        self.form_field = form_field


class ItemConfigDialog(tksimpledialog.Dialog, object):
    def __init__(self, master, item_fields, item_config=None, **kwargs):
        self.item_fields = item_fields
        self.item_config = item_config

        for ff in self.item_fields.values():
            ff.reset()

        super(ItemConfigDialog, self).__init__(master, **kwargs)

    def body(self, master):

        if self.item_config is not None:
            for k, ff in self.item_fields.items():
                if k in self.item_config:
                    ff.from_string(self.item_config[k])

        self.item_frame = form.FormFrame(master)
        self.item_frame.grid_fields(self.item_fields.values(), False)
        self.item_frame.grid(row=0, sticky='nsew')

        self.result = None

    def apply(self):
        self.result = collections.OrderedDict()

        for k, ff in self.item_fields.items():
            self.result[k] = ff.to_string()


class ListManager(tk.Frame):
    def __init__(self, master, columns_def, add=None, remove=None, edit=None):
        tk.Frame.__init__(self, master)

        self.master = master
        self.columns_def = columns_def

        self.fields = collections.OrderedDict([
            (k, cd.form_field) for k, cd in self.columns_def.items()
        ])

        self.items = None

        column_ids = [k for k in self.columns_def.keys() if k != '#0']

        self.tree = ttk.Treeview(self, columns=column_ids,
                                 selectmode='browse')

        ysb = ttk.Scrollbar(self, orient='vertical', command=self.tree.yview)
        xsb = ttk.Scrollbar(self, orient='horizontal', command=self.tree.xview)
        self.tree.configure(yscroll=ysb.set, xscroll=xsb.set)

        for c in ['#0'] + column_ids:
            cd = self.columns_def[c]
            self.tree.heading(c, text=cd.text, anchor=cd.anchor)

        self.tree.grid(row=0, column=0, sticky='nsew')
        ysb.grid(row=0, column=1, sticky='ns')
        xsb.grid(row=1, column=0, sticky='ew')

        butbox = tk.Frame(self)
        butbox.grid(row=0, column=2, sticky='ns')

        self.newbut = None
        self.removebut = None
        self.editbut = None

        if add is not None:
            self.newbut = tk.Button(butbox, text='Add', command=self.add)
            self.newbut.grid(row=0, column=0, sticky='ew')

        if remove is not None:
            self.removebut = tk.Button(butbox, text='Remove', command=self.remove,
                                       state=tk.DISABLED)
            self.removebut.grid(row=1, column=0, sticky='ew')

        if edit is not None:
            self.editbut = tk.Button(butbox, text='Edit', command=self.edit,
                                     state=tk.DISABLED)
            self.editbut.grid(row=2, column=0, sticky='ew')

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.grid(sticky='nsew')

        self.tree.bind('<<TreeviewSelect>>', self.selchanged)

    def populate(self, rows):
        for child in self.tree.get_children():
            self.tree.delete(child)

        self.items = collections.OrderedDict([(v['#0'], v) for v in rows])

        for v in rows:
            rows = [v[c] for c in self.columns_def if c != '#0' and c != 'iid']
            iid = v.get('iid', v['#0'])
            root_node = self.tree.insert('', 'end', iid=iid, text=v['#0'],
                                         open=True, values=rows)

    def add(self):
        n = ItemConfigDialog(self.master, self.fields)
        new = n.result
        if new is None:
            return

        name = new['#0']
        self.items[name] = new
        rows = [v for k, v in self.items[name].items() if k != '#0']
        self.tree.insert('', 'end', iid=name, text=name, values=rows)

    def remove(self):
        selected = self.tree.selection()[0]

        del self.items[selected]

        self.tree.selection_set('')
        self.tree.delete(selected)

    def edit(self):
        selected = self.tree.selection()[0]
        item = self.items[selected]

        icd = ItemConfigDialog(self.master, self.fields, item)

        # update list display
        self.items[selected] = icd.result
        rows = [v for k, v in self.items[selected].items() if k != '#0']
        self.tree.item(selected, text=icd.result['#0'], values=rows)

    def selchanged(self, event):
        buts = [b for b in [self.editbut, self.removebut] if b is not None]
        if self.tree.selection():
            for b in buts:
                b.config(state=tk.NORMAL)
        else:
            for b in buts:
                b.config(state=tk.DISABLED)


class List(object):
    def __init__(self, column_defs, rows = []):
        self.column_defs = column_defs
        self.rows = rows

    def get_widget(self, master):
        lm = ListManager(master, self.column_defs)
        lm.populate(self.rows)

        return lm


if __name__ == '__main__':
    import form

    root = tk.Tk()

    columns_def = collections.OrderedDict([
        ('#0', ColumnDef('#0', 'name', form_field=form.TextField('name:'))),
        ('col1', ColumnDef('col1', 'column 1', form_field=form.TextField('column 1 text:'))),
        ('col2', ColumnDef('col2', 'column 2', form_field=form.IntegerField('column 2 integer:'))),
    ])

    lm = ListManager(root, columns_def, add=True, remove=True,
                     edit=True)

    rows = [
        {'#0': 'value 1', 'col1': 'v1c1', 'col2': 1},
        {'#0': 'value 2', 'col1': 'v2c1', 'col2': 2},
    ]

    lm.populate(rows)




    root.mainloop()



import os
import errno


class BrocoliError(Exception):
    def __init__(self, exception):
        self.exception = exception

    def __str__(self):
        return type(self.exception).__name__ + ': ' + str(self.exception)


class ConnectionError(BrocoliError):
    pass

class FileNotFoundError(BrocoliError):
    pass


def ioerror(no):
    return IOError(no, os.strerror(no))




#!/usr/bin/env python

import six
from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_tksimpledialog as tksimpledialog
from six.moves import tkinter_ttk as ttk

import argparse
import logging
import sys
import os

from . import config
from . treewidget import TreeWidget
from . import catalog
from . import preferences

# Brocoli version string
__version__ = '0.1.0'


class ConnectionSwitcher:
    """
    Provides a method that can be used to switch to a particular connection
    """
    def __init__(self, app, conn_name):
        self.app = app
        self.conn_name = conn_name

    def switch(self):
        self.app.set_connection(self.conn_name)


class SwitcherSubmenu:
    """
    A menu containing a list of connections to switch to
    """
    def __init__(self, connection_menu, app):
        self.app = app
        self.menu = connection_menu
        self.menu_name = 'Switch connection'

        self.menu.add_cascade(label=self.menu_name, menu=None)

    def populate(self):
        index = self.menu.index(self.menu_name)
        self.menu.delete(index)

        switch_menu = tk.Menu(self.menu, tearoff=False)

        for c in self.app.cfg.connection_names():
            cs = ConnectionSwitcher(self.app, c)
            switch_menu.add_command(label=c, command=cs.switch)

        self.menu.insert_cascade(index, label=self.menu_name,
                                 menu=switch_menu)


class BrocoliApplication(object):
    def __init__(self, cfg, connection_name):
        # run Tk
        self.root = tk.Tk()

        self.cfg = cfg

        # get connection
        self.cat, root_path = cfg.connection(connection_name)

        self.path = root_path

        # create menus
        self.menubar = tk.Menu(self.root)

        self.connection_menu = tk.Menu(self.menubar, tearoff=False)

        self.menubar.add_cascade(label='Settings', menu=self.connection_menu)

        self.menubar.add_command(label="Quit!", command=self.root.quit)

        self.root.config(menu=self.menubar)

        # main window tree view, populate connection menu
        self.tree_widget = TreeWidget(self.root, self.path)
        self.tree_widget.grid(sticky='nsew')

        self.set_display_columns()

        self.connection_menu.add_command(label="New connection",
                                         command=self.new_connection)

        self.ss = SwitcherSubmenu(self.connection_menu, self)
        self.ss.populate()

        self.connection_menu.add_command(label="Preferences",
                                         command=self.open_preferences)

        # layout
        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)

        self.set_connection(connection_name)

    def run(self):
        self.root.mainloop()

    def set_display_columns(self):
        dcols = self.cfg[config.SETTINGS].get('display_columns', None)
        if dcols is not None:
            dcols = dcols.split(',')

        self.tree_widget.set_display_columns(dcols)

    def new_connection(self):
        dialog = preferences.ConnectionConfigDialog(self.root)

        if dialog.result is None:
            return

        config.save_config(dialog.to_config_dict(), update=True)
        self.cfg = config.load_config()
        self.ss.populate()

    def open_preferences(self):
        prefs = preferences.Preferences(self.root)
        if prefs.changed:
            self.cfg = prefs.cfg
            self.ss.populate()

            self.set_display_columns()

    def set_connection(self, connection_name):
        conn, path = self.cfg.connection(connection_name)
        self.tree_widget.set_connection(conn, path)


def main():
    """
    Main function for the Brocoli application. Handles command line arguments
    and loads configuration before launching the application window
    """

    parser = argparse.ArgumentParser(description='Browse catalog')
    parser.add_argument('--connection', metavar='CONNECTION',
                        default=None, help='use [connection:CONNECTION] '
                        'section in configuration file')

    args = parser.parse_args()

    cfg = config.load_config()

    app = BrocoliApplication(cfg, args.connection)

    app.run()

if __name__ == '__main__':
    main()



from . import catalog
from . import irodscatalog

from six.moves import configparser

import os
import os.path
import stat
import tempfile
import collections

# default config location
default_config_filename = os.path.join(os.path.expanduser('~'),
                                       '.brocoli.ini')

# available catalogs
catalog_dict = {
    'os': catalog.OSCatalog,
    'irods3': irodscatalog.iRODSCatalog,
}

catalog_types = list(catalog_dict.keys())

SETTINGS = 'SETTINGS'


class Config(collections.OrderedDict):
    """
    Holds the brocoli configuration loaded from the config file
    """

    def connection(self, name=None):
        name = name or self[SETTINGS]['default_connection']

        conn = self['connection:' + name]

        cat = None
        catalog_type = conn['catalog_type']
        if catalog_type == 'os':
            cat = lambda master: catalog.OSCatalog()
        elif catalog_type == 'irods3':
            cat = irodscatalog.irods3_catalog_from_config(conn)

        return cat, conn['root_path']

    def connection_names(self):
        return [k.split(':', 1)[1] for k in self if k.startswith('connection:')]


def bootstrap_config_parser(config):
    """
    Fills ConfigParser object with minimal values
    """
    config.add_section(SETTINGS)
    config.set(SETTINGS, 'default_connection', 'default')
    config.add_section('connection:default')
    config.set('connection:default', 'catalog_type', 'os')
    config.set('connection:default', 'root_path', tempfile.gettempdir())


def load_config(filename=None):
    """
    Loads configuration from file
    """
    filename = filename or default_config_filename

    config = configparser.RawConfigParser()

    if os.path.exists(filename):
        config.read(filename)
    else:
        bootstrap_config_parser(config)

    ret = Config()

    for section in config.sections():
        ret[section] = collections.OrderedDict(config.items(section))

    return ret


def save_config(config_dict, filename=None, update=False):
    """
    Writes configuration to a file
    """
    filename = filename or default_config_filename

    config = configparser.RawConfigParser()

    if update:
        if os.path.exists(filename):
            config.read(filename)
        else:
            bootstrap_config_parser(config)

    for section, section_content in config_dict.items():
        if not config.has_section(section):
            config.add_section(section)

        for option, option_value in section_content.items():
            config.set(section, option, option_value)

    with open(filename, 'wt') as f:
        os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)
        config.write(f)



"""
Implements iRODS Catalog object and methods
"""

from . import catalog
from . import form
from . import exceptions
from . listmanager import ColumnDef, List

import re
import os
import hashlib
import collections

from six import print_
from six.moves import tkinter as tk

import irods
from irods.session import iRODSSession
from irods import password_obfuscation
from irods.manager.data_object_manager import DataObjectManager
from irods.manager.collection_manager import CollectionManager
from irods.models import DataObject, Collection
import irods.keywords as kw
import irods.exception

_getuid = None
if hasattr(os, 'getuid'):
    _getuid = os.getuid
else:
    # generate a fake uid on systems that lacks os.getuid()
    _getuid = lambda: int('0x' + hashlib.md5(os.getlogin().encode()).hexdigest(), 16) % 10000


def parse_env3(path):
    """
    parse iRODS v3 iCommands environment files
    """

    pat = ("^\s*(?P<name>\w+)" + "(\s+|\s*=\s*)[\'\"]?(?P<value1>[^\'\"\n]*)" +
           "[\'\"]?\s*$")
    envre = re.compile(pat)

    ret = {}

    with open(path, 'r') as f:
        for l in f.readlines():
            m = envre.match(l)
            if m:
                ret[m.group("name")] = m.group("value1")

    return ret


def local_tree_stats(dirs):
    """
    Gathers stats (number of files and cumulated size) of sub-trees on a local
    directories list
    """
    total_nfiles = 0
    total_size = 0

    for d in dirs:
        nfiles = 0
        size = 0

        for root, dirs, files in os.walk(d):
            nfiles += len(files)
            size += sum(os.path.getsize(os.path.join(root, name))
                        for name in files)

        total_nfiles += nfiles
        total_size += size

    return total_nfiles, total_size


def local_files_stats(files):
    """
    Computes stats (number of files and cumulated size) on a file list
    """
    return len(files), sum(os.path.getsize(f) for f in files)


def translate_exceptions(method):
    """
    Method decorator that translates iRODS to Brocoli exceptions
    """
    def method_wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            raise exceptions.ConnectionError(e)
        except irods.exception.CAT_UNKNOWN_COLLECTION as e:
            raise exceptions.FileNotFoundError(e)

    return method_wrapper


class iRODSCatalog(catalog.Catalog):
    """
    A Catalog for connectiong to iRODS
    """
    @classmethod
    def encode(cls, s):
        return password_obfuscation.encode(s, _getuid())

    @classmethod
    def decode(cls, s):
        return password_obfuscation.decode(s, _getuid())

    def __init__(self, host, port, user, zone, scrambled_password):
        try:
            password = iRODSCatalog.decode(scrambled_password)
            self.session = iRODSSession(host=host, port=port, user=user,
                                        password=password, zone=zone)
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            raise exceptions.ConnectionError(e)

        self.dom = self.session.data_objects
        self.cm = self.session.collections
        self.am = self.session.permissions

    def splitname(self, path):
        return path.rsplit('/', 1)

    def basename(self, path):
        _, basename = self.splitname(path)
        return basename

    def dirname(self, path):
        dirname, _ = self.splitname(path)
        return dirname

    @translate_exceptions
    def lstat(self, path):
        if self.isdir(path):
            return self.lstat_dir(path)
        return self.lstat_file(path)

    def lstat_dir(self, path):
        q = self.session.query(Collection.owner_name)
        q = q.filter(Collection.name == path)

        r = q.one()

        ret = {
            'user': r[Collection.owner_name],
            'size': '',
            'mtime': '',
            'nreplicas': '',
        }

        return ret

    def lstat_file(self, path):
        dirname, basename = self.splitname(path)
        q = self.session.query(DataObject.owner_name, DataObject.size,
                               DataObject.modify_time)
        q = q.filter(Collection.name == dirname)
        q = q.filter(DataObject.name == basename)

        replicas = q.all()
        if len(replicas) < 1:
            # no replica
            raise exceptions.ioerror(exceptions.errno.ENOENT)

        ret = {'nreplicas': len(replicas)}
        minsize = None
        maxsize = 0
        for r in replicas:
            ret['user'] = r[DataObject.owner_name]
            ret['mtime'] = r[DataObject.modify_time]
            size = r[DataObject.size]
            if minsize is None or size < minsize:
                minsize = size
            if size > maxsize:
                maxsize = size

        if maxsize != minsize:
            ret['size'] = '{}-{}'.format(minsize, maxsize)
        else:
            ret['size'] = str(minsize)

        return ret

    @translate_exceptions
    def listdir(self, path):
        q = self.session.query(DataObject.name).filter(Collection.name == path)
        files = [r[DataObject.name] for r in q.all()]

        q = self.session.query(Collection.name) \
            .filter(Collection.parent_name == path)
        colls = [self.basename(c[Collection.name]) for c in q.all()]
        return colls + files

    @translate_exceptions
    def isdir(self, path):
        q = self.session.query(Collection.id).filter(Collection.name == path)

        try:
            q.one()
        except irods.exception.NoResultFound:
            return False

        return True

    def join(self, *args):
        return '/'.join(args)

    @translate_exceptions
    def download_files(self, pathlist, destdir):
        options = {kw.FORCE_FLAG_KW: ''}

        number = len(pathlist)
        i = 0
        for p in pathlist:
            self.dom._download(p, destdir, **options)
            i += 1
            yield i, number

    def _download_coll(self, coll, destdir):
        destdir = os.path.join(destdir, coll.name)
        try:
            os.makedirs(destdir)
        except OSError:
            if not os.path.isdir(destdir):
                raise

        pathlist = [dobj.path for dobj in list(coll.data_objects)]
        for y in self.download_files(pathlist, destdir):
            yield y

        for subcoll in coll.subcollections:
            for y in self._download_coll(subcoll, destdir):
                yield y

    @translate_exceptions
    def download_directories(self, pathlist, destdir):
        for p in pathlist:
            coll = self.cm.get(p)
            for y in self._download_coll(coll, destdir):
                yield y

    @translate_exceptions
    def upload_files(self, files, path):
        nfiles, size = local_files_stats(files)

        completed = 0
        for s in self._upload_files(files, path):
            completed += s
            yield completed, size

    def _upload_files(self, files, path):
        def local_file_md5(filename):
            m = hashlib.md5()
            with open(filename, 'rb') as f:
                m.update(f.read())

            return m.hexdigest()

        if not path.endswith('/'):
            path = path + '/'

        options = {
            kw.VERIFY_CHKSUM_KW: '',
            kw.ALL_KW: '',
            kw.UPDATE_REPL_KW: '',
        }

        for f in files:
            basename = os.path.basename(f)
            irods_path = path + basename
            print_('put', f, path, irods_path)

            options[kw.VERIFY_CHKSUM_KW] = local_file_md5(f)

            self.dom.put(f, irods_path, **options)

            yield os.path.getsize(f)

    def _upload_dir(self, dir_, path):
        files = []
        subdirs = []
        for name in os.listdir(dir_):
            abspath = os.path.join(dir_, name)
            if os.path.isdir(abspath):
                subdirs.append((abspath, name))
            else:
                files.append(abspath)

        for y in self._upload_files(files, path):
            yield y

        for abspath, name in subdirs:
            cpath = self.join(path, name)
            try:
                coll = self.cm.create(cpath)
                remove = False
            except irods.exception.CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME:
                pass

            for y in self._upload_dir(abspath, cpath):
                yield y

    @translate_exceptions
    def upload_directories(self, dirs, path):
        nfiles, size = local_tree_stats(dirs)

        completed = 0
        for d in dirs:
            name = os.path.basename(d)
            cpath = self.join(path, name)

            try:
                self.cm.create(cpath)
            except irods.exception.CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME:
                pass

            for s in self._upload_dir(d, cpath):
                completed += s
                yield completed, size

    @translate_exceptions
    def delete_files(self, files):
        number = len(files)

        i = 0
        for f in files:
            self.dom.unlink(f, force=True)
            i += 1
            yield i, number

    @translate_exceptions
    def delete_directories(self, directories):
        number = len(directories)

        i = 0
        for d in directories:
            self.cm.remove(d, recurse=True, force=True)
            i += 1
            yield i, number

    @translate_exceptions
    def mkdir(self, path):
        self.cm.create(path)

    def __acls_from_object(self, obj):
        access = self.am.get(obj)

        acls = [a.__dict__.copy() for a in access]

        for a in acls:
            # build a unique ttk.TreeWidget iid
            a['iid'] = '#'.join([a[t] for t in ['user_name', 'user_zone',
                                                'access_name']])
            a['#0'] = a['user_name']

        return List(iRODSCatalog.acls_def(), acls)

    def __metadata_from_object(self, obj):
        metadata = [md.__dict__.copy() for md in obj.metadata.items()]

        for md in metadata:
            md['iid'] = md['avu_id']
            md['#0'] = md['name']

        return List(iRODSCatalog.metadata_def(), metadata)

    @translate_exceptions
    def directory_properties(self, path):
        co = self.cm.get(path)
        acls_list = self.__acls_from_object(co)
        metadata_list = self.__metadata_from_object(co)

        return collections.OrderedDict([
            ('Permissions', acls_list),
            ('Metadata', metadata_list),
        ])


    @translate_exceptions
    def file_properties(self, path):
        do = self.dom.get(path)
        replicas = [r.__dict__.copy() for r in do.replicas]
        for r in replicas:
            # set row title for ListManager use
            r['#0'] = r['number']

        replicas_list = List(iRODSCatalog.replicas_def(), replicas)

        acls_list = self.__acls_from_object(do)
        metadata_list = self.__metadata_from_object(do)

        return collections.OrderedDict([
            ('Replicas', replicas_list),
            ('Permissions', acls_list),
            ('Metadata', metadata_list),
        ])

    @classmethod
    def replicas_def(cls):
        repl_num = ColumnDef('#0', 'Number',
                             form_field=form.IntegerField('Replica number:',
                                                          -1))
        resc = ColumnDef('resource_name', 'Resource',
                         form_field=form.TextField('Resource name:'))
        path = ColumnDef('path', 'Path',
                         form_field=form.TextField('Replica path:'))
        status = ColumnDef('status', 'Status',
                           form_field=form.TextField('Replica status:'))
        checksum = ColumnDef('checksum', 'Checksum',
                           form_field=form.TextField('Replica checksum:'))

        cols = [repl_num, resc, status, checksum, path]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def acls_def(cls):
        user = ColumnDef('#0', 'User',
                             form_field=form.TextField('User:'))
        zone = ColumnDef('user_zone', 'Zone',
                         form_field=form.TextField('User zone:'))
        type = ColumnDef('access_name', 'Access type',
                         form_field=form.TextField('Acces type:'))


        cols = [user, zone, type]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def metadata_def(cls):
        name = ColumnDef('#0', 'Name',
                             form_field=form.IntegerField('Metadata name:'))
        value = ColumnDef('value', 'Value',
                             form_field=form.IntegerField('Metadata value:'))
        unit = ColumnDef('units', 'Unit',
                             form_field=form.IntegerField('Metadata unit:'))

        cols = [name, value, unit]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def config_fields(cls):

        tags = ['inline_config']

        return collections.OrderedDict([
            ('use_irods_env', form.BooleanField('Use irods environment file',
                                                disables_tags=tags)),
            ('host', form.HostnameField('iRODS host:', tags=tags)),
            ('port', form.IntegerField('iRODS port:', '1247', tags=tags)),
            ('zone', form.TextField('iRODS zone:', tags=tags)),
            ('user_name', form.TextField('iRODS user name:', tags=tags)),
            ('store_password', form.BooleanField('Remember password',
                                                 enables_tags=['password'],
                                                 tags=tags)),
            ('password', form.PasswordField('iRODS password:',
                                            encode=cls.encode,
                                            decode=cls.decode,
                                            tags=tags + ['password'])),
        ])


def irods3_catalog_from_envfile(envfile):
    """
    Creates an iRODSCatalog from a iRODS v3 configuration file (like
    "~/.irods/.irodsEnv")
    """
    env3 = parse_env3(envfile)

    host = env3['irodsHost']
    port = int(env3.get('irodsPort', '1247'))
    user = env3['irodsUserName']
    zone = env3['irodsZone']
    pwdfile = env3['irodsAuthFileName']

    with open(pwdfile, 'r') as f:
        scrambled_password = f.read().strip()

    return iRODSCatalog(host, port, user, zone, scrambled_password)


def irods3_catalog_from_config(cfg):
    """
    Creates an iRODSCatalog from configuration
    """
    use_env = False
    if cfg['use_irods_env'].lower() in ['1', 'yes', 'on', 'true']:
        use_env = True
    elif cfg['use_irods_env'].lower() in ['0', 'no', 'off', 'false']:
        use_env = False
    else:
        msg = 'invalid irods_use_env value: {}'.format(cfg['use_irods_env'])
        raise ValueError(msg)

    if use_env:
        envfile = os.path.join(os.path.expanduser('~'), '.irods', '.irodsEnv')
        return lambda master: irods3_catalog_from_envfile(envfile)

    host = cfg['host']
    port = cfg['port']
    user = cfg['user_name']
    zone = cfg['zone']
    store_password = cfg['store_password']
    scrambled_password = None
    if store_password.lower() in ['1', 'yes', 'on', 'true']:
        scrambled_password = cfg['password']
        return lambda master: iRODSCatalog(host, port, user, zone,
                                           scrambled_password)
    else:
        def ask_password(master):
            tl = tk.Toplevel(master)
            tl.transient(master)

            ff = form.FormFrame(tl)
            pf = form.PasswordField('enter iRODS password:',
                                    return_cb=tl.destroy)
            ff.grid_fields([pf])
            ff.pack()

            tl.wait_window()

            scrambled_password = iRODSCatalog.encode(pf.to_string())

            return iRODSCatalog(host, port, user, zone, scrambled_password)

        return ask_password



import os.path

from catabrowse.irodscatalog import iRODSCatalog, make_irods3_catalog

from irods.models import DataObject, Collection
from irods.exception import *

from six import print_

ic = make_irods3_catalog('/home/pigay/.irods/.irodsEnv')

print_(ic)
print_(ic.dom)
print_(ic.cm)

coll = ic.session.collections.get('/MCIA/home/pigay/testdir')

print_(coll)
print_(coll.path)
print_(coll.metadata.keys())

q = ic.session.query(DataObject.name, DataObject.owner_name, DataObject.size, DataObject.modify_time).filter(Collection.id == coll.id)

#for r in q.all():
#    print_(r)

print_('coucou')
q = ic.session.query(DataObject.name, DataObject.owner_name, DataObject.size, DataObject.modify_time, Collection.name).filter(Collection.name == coll.path).filter(DataObject.name == 'unison.log')
for r in q.all():
    print_(r)

print_(ic.lstat('/MCIA/home/pigay/testdir/unison.log'))

print_(ic.listdir('/MCIA/home/pigay/testdir'))
print_(ic.isdir('/MCIA/home/pigay/testdir'))
print_(ic.isdir('/MCIA/home/pigay/testdir/unison.log'))

print_(ic.join('a', 'b', 'c'))

ic.download_files(['/MCIA/home/pigay/testdir/unison.log'], '/tmp')

testcoll = '/MCIA/home/pigay/testdir/testcatabrowse'

try:
    ic.mkdir(testcoll)
except CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME:
    print_('collection already present. mkdir not tested')

ic.upload_files([os.path.abspath('catabrowse.geany'), os.path.abspath('.gitignore')], testcoll)

ic.download_files([ic.join(testcoll, 'catabrowse.geany'), ic.join(testcoll, '.gitignore')], '/tmp')

ic.delete_files([ic.join(testcoll, '.gitignore'), ])


ic.upload_directories([os.path.abspath('catabrowse')], testcoll)

ic.download_directories([ic.join(testcoll, 'catabrowse'), ], '/tmp')

ic.delete_directories([ic.join(testcoll, 'catabrowse'), ])



#! /usr/bin/env python

# ideas stolen from https://github.com/jgehrcke/python-cmdline-bootstrap

import re

from setuptools import setup, find_packages

version = re.search(
    '^__version__\s*=\s*\'(.*)\'',
    open('brocoli/brocoli.py').read(),
    re.M
    ).group(1)


with open("README.rst", "rb") as f:
    long_descr = f.read().decode("utf-8")

setup(name='brocoli',
      version=version,
      description='Browse Collections for iRODS',
      long_description=long_descr,
      author='Pierre Gay',
      author_email='pierre.gay@u-bordeaux.fr',
      url='https://github.com/mesocentre-mcia/brocoli',
      packages=find_packages('.', exclude=['*.tests']),
      entry_points = {
        "console_scripts": ['brocoli = brocoli.brocoli:main']
        },
      python_requires='>=2.7',
      keywords=['irods', 'tkinter'],
      classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
        'Topic :: Desktop Environment :: File Managers',
      ],
      install_requires=[
      'six>=1.10.0',
      'python-irodsclient>=0.7.0',
      ]
     )
